# coding=utf-8
from __future__ import unicode_literals, division

import datetime

import pandas as pd
import os

import struct
from jotdx.reader.base_reader import TdxFileNotFoundException
from jotdx.reader.base_reader import BaseReader

"""
读取通达信数据
"""


class TdxExHqDailyBarReader(BaseReader):

    def __init__(self, vipdoc_path=None):
        self.vipdoc_path = vipdoc_path

    def parse_data_by_file(self, fname):
        if not os.path.isfile(fname):
            raise TdxFileNotFoundException('no tdx kline data, pleaes check path %s', fname)

        with open(fname, 'rb') as f:
            content = f.read()
            return self.unpack_records('<hhffffIIf', content)

        return []

    def get_df(self, code_or_file):
        # 只传入了一个参数
        data = [self._df_convert(row) for row in self.parse_data_by_file(code_or_file)]

        df = pd.DataFrame(
            data=data,
            columns=('date', 'open', 'high', 'low', 'close',
                     'amount', 'volume', 'jiesuan', 'hk_stock_amount')
        )
        # df.index = pd.to_datetime(df.date)

        # TODO 通达信这里还是有点问题,
        # (1) 按照现在这么写, 如果是星期一, 那么0点前的数据推到了星期日,
        #  0点后到9点开盘前应该是上周五晚上的行情, 算作星期一凌晨
        # 暂时不影响回测, 这里有个小坑注意
        ddelay = datetime.timedelta(days=1)
        df['date'] = pd.to_datetime(df['date']).apply(lambda x: x if 0 <= x.hour < 16 else x - ddelay)

        df = df.set_index('date')

        return df

    def _df_convert(self, row):
        t_date = row[0]
        year = int(t_date / 2048 + 2036)
        year = (year - 32) if (year > datetime.datetime.now().year) else year

        month = int(t_date % 2048 / 100)
        day = t_date % 2048 % 100

        t_time = row[1]
        hour = int(t_time / 60)
        min = t_time % 60

        datetimep = datetime.datetime(year=year, month=month, day=day, hour=hour, minute=min)

        offset = 1
        (hk_stock_amount,) = struct.unpack('<f', struct.pack('<I', row[5 + offset]))
        new_row = (
            datetimep,
            row[1 + offset],
            row[2 + offset],
            row[3 + offset],
            row[4 + offset],
            row[5 + offset],
            row[6 + offset],
            row[7 + offset],
            hk_stock_amount
        )

        return new_row


if __name__ == '__main__':
    tdx_reader = TdxExHqDailyBarReader()
    try:
        print(tdx_reader.get_df("/Users/rainx/tmp/vipdoc/ds/29#A1801.day"))
        # print(tdx_reader.get_df("/Volumes/share/transfer/76#AG200.day"))

    except TdxFileNotFoundException as e:
        pass



# coding=utf-8

from jotdx.parser.base import BaseParser
from jotdx.helper import get_datetime, get_volume, get_price
from collections import OrderedDict
import struct
import six
import sys


class GetReportFile(BaseParser):
    def setParams(self, filename, offset=0):
        pkg = bytearray.fromhex(u'0C 12 34 00 00 00')
        # Fom DTGear request.py file
        node_size = 0x7530
        raw_data = struct.pack(r"<H2I100s", 0x06B9,
                               offset, node_size, filename.encode("utf-8"))
        raw_data_len = struct.calcsize(r"<H2I100s")
        pkg.extend(struct.pack(u"<HH{}s".format(raw_data_len),
                               raw_data_len, raw_data_len, raw_data))
        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        (chunksize, ) = struct.unpack("<I", body_buf[:4])

        if chunksize > 0:
            return {
                "chunksize": chunksize,
                "chunkdata":  body_buf[4:]
            }
        else:
            return {
                "chunksize": 0
            }


if __name__ == "__main__":
    from jotdx.hq import TdxHq_API
    api = TdxHq_API()
    api.need_setup = False
    # calc.tdx.com.cn, calc2.tdx.com.cn
    with api.connect(ip="120.76.152.87"):
        # response = api.get_report_file(r"tdxfin/gpcw19980630.zip", 386003)
        content = api.get_report_file_by_size("tdxfin/gpcw.txt")
        # content = api.get_report_file_by_size("tdxfin/gpcw19980630.zip", 386073)
        print(content)



# coding=utf-8

from jotdx.parser.base import BaseParser
import struct


# bytearray(b'TDX_DS\x00\x00\x00\x00\x00\x1f\xdc\x00\x00\x01\x00\x00\x00=\x9c\x00\x00t\x00\x00\x00\x00\x00\x00\x00')
class GetInstrumentCount(BaseParser):

    def setup(self):
        self.send_pkg = bytearray.fromhex("01 03 48 66 00 01 02 00 02 00 f0 23")

    def parseResponse(self, body_buf):
        pos = 0
        (num,) = struct.unpack("<I", body_buf[19: 19 + 4])
        return num



# -*- coding: utf-8 -*-
# @Author  : BoPo
# @Time    : 2021/10/11 17:28
# @Function:
import json

import httpx
import pandas as pd
from tenacity import stop_after_attempt, wait_fixed, retry

from jotdx.consts import return_last_value


@retry(wait=wait_fixed(2), retry_error_callback=return_last_value, stop=stop_after_attempt(5))
def fq_factor(method: str, symbol: str) -> pd.DataFrame:
    zh_sina_a_stock_hfq_url = "https://finance.sina.com.cn/realstock/company/{}/hfq.js"
    zh_sina_a_stock_qfq_url = "https://finance.sina.com.cn/realstock/company/{}/qfq.js"

    client = httpx.Client(verify=False)

    if method == "hfq":
        res = client.get(zh_sina_a_stock_hfq_url.format(symbol))
        hfq_factor_df = pd.DataFrame(json.loads(res.text.split("=")[1].split("\n")[0])["data"])

        if hfq_factor_df.shape[0] == 0:
            raise ValueError("sina hfq factor not available")

        hfq_factor_df.columns = ["date", "hfq_factor"]
        hfq_factor_df.index = pd.to_datetime(hfq_factor_df.date)

        del hfq_factor_df["date"]

        hfq_factor_df.reset_index(inplace=True)
        # hfq_factor_df = hfq_factor_df.set_index('date')

        return hfq_factor_df
    else:
        res = client.get(zh_sina_a_stock_qfq_url.format(symbol))
        qfq_factor_df = pd.DataFrame(json.loads(res.text.split("=")[1].split("\n")[0])["data"])

        if qfq_factor_df.shape[0] == 0:
            raise ValueError("sina hfq factor not available")

        qfq_factor_df.columns = ["date", "qfq_factor"]
        qfq_factor_df.index = pd.to_datetime(qfq_factor_df.date)

        del qfq_factor_df["date"]

        qfq_factor_df.reset_index(inplace=True)
        # qfq_factor_df = qfq_factor_df.set_index('date')

        return qfq_factor_df


def to_adjust(temp_df, symbol=None, adjust=None):
    # zh_sina_a_stock_hfq_url = "https://finance.sina.com.cn/realstock/company/{}/hfq.js"
    # zh_sina_a_stock_qfq_url = "https://finance.sina.com.cn/realstock/company/{}/qfq.js"

    temp_df['volume'] = temp_df['vol']
    temp_df['date'] = pd.to_datetime(temp_df[['year', 'month', 'day']])
    temp_df = temp_df.set_index('date')

    if adjust == "hfq":
        # res = requests.get(zh_sina_a_stock_hfq_url.format(symbol))
        # hfq_factor_df = pd.DataFrame(eval(res.text.split("=")[1].split("\n")[0])["data"])
        # hfq_factor_df.columns = ["date", "hfq_factor"]
        # hfq_factor_df.index = pd.to_datetime(hfq_factor_df.date)

        hfq_factor_df = fq_factor(symbol=symbol, method=adjust)
        del hfq_factor_df["date"]

        temp_df = pd.merge(temp_df, hfq_factor_df, left_index=True, right_index=True, how="outer")
        temp_df.fillna(method="ffill", inplace=True)
        temp_df = temp_df.astype(float)
        temp_df.dropna(inplace=True)
        temp_df.drop_duplicates(subset=["open", "high", "low", "close", "volume"], inplace=True)

        for field in ["open", "high", "low", "close"]:
            temp_df[field] = round(temp_df[field] * temp_df["hfq_factor"], 2)

        temp_df = temp_df.iloc[:, :-1]
        # temp_df = temp_df[start_date:end_date]

        temp_df.dropna(inplace=True)
        temp_df.reset_index(inplace=True)

        return temp_df

    if adjust == "qfq":
        # res = requests.get(zh_sina_a_stock_qfq_url.format(symbol))
        # qfq_factor_df = pd.DataFrame(eval(res.text.split("=")[1].split("\n")[0])["data"])
        # qfq_factor_df.columns = ["date", "qfq_factor"]
        # qfq_factor_df.index = pd.to_datetime(qfq_factor_df.date)
        qfq_factor_df = fq_factor(symbol=symbol, method=adjust)
        qfq_factor_df = qfq_factor_df.set_index('date')
        # del qfq_factor_df["date"]

        temp_df = pd.merge(temp_df, qfq_factor_df, left_index=True, right_index=True, how="outer")
        temp_df.fillna(method="ffill", inplace=True)

        # temp_df = temp_df.astype(float)

        for field in ["open", "high", "low", "close", "volume", 'qfq_factor']:
            temp_df[field] = temp_df[field].astype(float)

        temp_df.dropna(inplace=True)
        temp_df.drop_duplicates(subset=["open", "high", "low", "close", "volume"], inplace=True)

        for field in ["open", "high", "low", "close"]:
            temp_df[field] = round(temp_df[field] / temp_df["qfq_factor"], 2)

        temp_df = temp_df.iloc[:, :-1]
        temp_df.dropna(inplace=True)
        temp_df.reset_index(inplace=True)

        print(temp_df)
        return temp_df

    return temp_df



import asyncio
import functools
import json
import socket
import time
from functools import partial

from jotdx.consts import CONFIG
from jotdx.consts import EX_HOSTS
from jotdx.consts import GP_HOSTS
from jotdx.consts import HQ_HOSTS
from jotdx.logger import logger
from jotdx.utils import get_config_path

hosts = {
    'HQ': [{'addr': hs[1], 'port': hs[2], 'time': 0, 'site': hs[0]} for hs in HQ_HOSTS],
    'EX': [{'addr': hs[1], 'port': hs[2], 'time': 0, 'site': hs[0]} for hs in EX_HOSTS],
    'GP': [{'addr': hs[1], 'port': hs[2], 'time': 0, 'site': hs[0]} for hs in GP_HOSTS],
}

results = {k: [] for k in hosts}


def callback(res, key):
    result = res.result()

    if result.get('time'):
        results[key].append(result)

    logger.debug('callback: {}', res.result())


def connect(proxy):
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(3)

        start = time.perf_counter()

        sock.connect((proxy.get('addr'), int(proxy.get('port'))))
        sock.close()

        proxy['time'] = (time.perf_counter() - start) * 1000

        logger.info('{addr}:{port} 验证通过，响应时间：{time} ms.'.format(**proxy))
    except socket.timeout as ex:
        logger.info('{addr},{port} time out.'.format(**proxy))
        proxy['time'] = None
    except ConnectionRefusedError as ex:
        logger.info('{addr},{port} 验证失败.'.format(**proxy))
        proxy['time'] = None
    finally:
        return proxy


async def verify(proxy):
    result = await asyncio.get_event_loop().run_in_executor(None, functools.partial(connect, proxy=proxy))
    return result


def Server(index=None, limit=5, console=False, sync=True):
    _hosts = hosts[index]

    def async_event():
        event = asyncio.get_event_loop()
        tasks = []

        while len(_hosts) > 0:
            task = event.create_task(verify(_hosts.pop(0)))
            task.add_done_callback(partial(callback, key=index))
            tasks.append(task)

        # event.is_closed()
        # event.is_running()
        event.run_until_complete(asyncio.wait(tasks))

    global results

    if sync:
        results[index] = [connect(proxy) for proxy in _hosts]
        results[index] = [x for x in results[index] if x.get('time')]
    else:
        async_event()

    server = results[index]

    # 结果按响应时间从小到大排序
    if console:
        from prettytable import PrettyTable

        server.sort(key=lambda item: item['time'])
        print('[√] 最优服务器:')

        t = PrettyTable(['Name', 'Addr', 'Port', 'Time'])
        t.align['Name'] = 'l'
        t.align['Addr'] = 'l'
        t.align['Port'] = 'l'
        t.align['Time'] = 'r'
        t.padding_width = 1

        for host in server[: int(limit)]:
            t.add_row(
                [
                    host['site'],
                    host['addr'],
                    host['port'],
                    '{:5.2f} ms'.format(host['time']),
                ]
            )

        print(t)

    return [(item['addr'], item['port']) for item in server]


def bestip(console=False, limit=5, sync=True) -> None:
    config_ = get_config_path('config.json')
    default = dict(CONFIG)

    for index in ['HQ', 'EX', 'GP']:
        try:
            data = Server(index=index, limit=limit, console=console, sync=sync)
            if data:
                default['BESTIP'][index] = data[0]
        except RuntimeError as ex:
            logger.error('请手动运行`python -m mootdx bestip`')
            break

    json.dump(default, open(config_, 'w', encoding='utf-8'), indent=2, ensure_ascii=False)


if __name__ == '__main__':
    bestip()



from jotdx.reader.daily_bar_reader import TdxDailyBarReader, TdxFileNotFoundException, TdxNotAssignVipdocPathException
from jotdx.reader.min_bar_reader import TdxMinBarReader
from jotdx.reader.lc_min_bar_reader import TdxLCMinBarReader
from jotdx.reader.exhq_daily_bar_reader import TdxExHqDailyBarReader
from jotdx.reader.gbbq_reader import GbbqReader
from jotdx.reader.block_reader import BlockReader
from jotdx.reader.block_reader import CustomerBlockReader
from jotdx.reader.history_financial_reader import HistoryFinancialReader
from jotdx.reader.mootdx_reader import Reader

__all__ = [
    'TdxDailyBarReader',
    'TdxFileNotFoundException',
    'TdxNotAssignVipdocPathException',
    'TdxMinBarReader',
    'TdxLCMinBarReader',
    'TdxExHqDailyBarReader',
    'GbbqReader',
    'BlockReader',
    'CustomerBlockReader',
    'HistoryFinancialReader',
    'Reader',
]



from jotdx import config
from jotdx.consts import EX_HOSTS
from jotdx.consts import GP_HOSTS
from jotdx.consts import HQ_HOSTS
from jotdx.server import Server
from jotdx.utils import get_config_path

__version__ = '0.1.4'



# coding=utf-8


from jotdx.parser.base import BaseParser
from collections import OrderedDict
import struct


class GetHistoryInstrumentBarsRange(BaseParser):
    def __init__(self, *args, **kvargs):
        self.seqid = 1
        BaseParser.__init__(self, *args, **kvargs)

    def setParams(self, market, code, date, date2):
        pkg = bytearray.fromhex('01')
        pkg.extend(struct.pack("<B", self.seqid))
        self.seqid = self.seqid + 1
        pkg.extend(bytearray.fromhex('38 92 00 01 16 00 16 00 0D 24'))
        code = code.encode("utf-8")
        # x =struct.pack("<B9s",  market, code)
        pkg.extend(struct.pack("<B9s", market, code))
        pkg.extend(bytearray.fromhex('07 00'))
        pkg.extend(struct.pack("<LL", date, date2))
        # print(hexdump.hexdump(pkg))
        self.send_pkg = pkg

    #

    def _parse_date(self, num):
        year = num // 2048 + 2004
        month = (num % 2048) // 100
        day = (num % 2048) % 100

        return year, month, day

    def _parse_time(self, num):
        return (num // 60), (num % 60)

    def parseResponse(self, body_buf):
        #        print('测试', body_buf)
        #        fileobj = open("a.bin", 'wb')  # make partfile
        #        fileobj.write(body_buf)  # write data into partfile
        #        fileobj.close()
        # print(hexdump.hexdump(body_buf[0:1024]))
        #        import zlib
        #        d=zlib.decompress(body_buf[16:])
        #        print(hexdump.hexdump(d))
        klines = []
        pos = 12

        # 算了，前面不解析了，没太大用
        # (market, code) = struct.unpack("<B9s", body_buf[0: 10]

        (ret_count,) = struct.unpack("H", body_buf[pos: pos + 2])
        pos = pos + 2
        # print(hexdump.hexdump(body_buf[20:52]))
        # print(hexdump.hexdump(body_buf[20: 20+ret_count*32]))
        # global raw_li
        print(ret_count)

        for i in range(ret_count):
            (d1, d2, open_price, high, low, close, position, trade, settlementprice) = struct.unpack("<HHffffIIf",
                                                                                                     body_buf[
                                                                                                     pos:pos + 32])
            # print(raw_li[0])
            pos = pos + 32
            # print(i)
            # pass
            # print(raw_li[i][0])
            year, month, day = self._parse_date(d1)
            hour, minute = self._parse_time(d2)
            #            print('%02d%02d%02d %02d %02d'%(year,month,day,hour,minute))
            #            print('%5.2f %5.2f %5.2f %5.2f %7d %7d %5.2f'%(open_price, high, low, close, position, trade, settlementprice))
            #            (open_price, high, low, close, position, trade, price) = struct.unpack("<ffffIIf", body_buf[pos: pos+28])
            #            pos += 28
            kline = OrderedDict([
                ("datetime", "%d-%02d-%02d %02d:%02d" % (year, month, day, hour, minute)),
                ("year", year),
                ("month", month),
                ("day", day),
                ("hour", hour),
                ("minute", minute),
                ("open", open_price),
                ("high", high),
                ("low", low),
                ("close", close),
                ("position", position),
                ("trade", trade),
                ("settlementprice", settlementprice)

            ])
            klines.append(kline)

        return klines


# 00000000  01 01 08 6A 01 01 16 00  16 00 FF 23 2F 49 46 4C   ...j.... ...#/IFL
# 00000010  30 00 F0 F4 94 13 07 00  01 00 00 00 00 00 F0 00   0....... ........

# 00000000: 01 01 08 6A 01 01 16 00  16 00 FF 23 4A 4E 56 44  ...j.......#JNVD
# 00000010: 41 00 C0 EC A3 13 07 00  01 00 00 00 00 00 C0 03  A...............

# 00000000  01 01 08 6A 01 01 16 00  16 00 FF 23 2F 49 46 31   ...j.... ...#/IF1
# 00000010  37 30 39 00 94 13 07 00  01 00 00 00 00 00 F0 00   709..... ........

if __name__ == '__main__':
    import pprint
    from jotdx.exhq import TdxExHq_API

    api = TdxExHq_API()
    with api.connect('61.152.107.141', 7727):
        x = api.to_df(api.get_history_instrument_bars_range(74, "BABA", 20170613, 20170620))
        pprint.pprint(x.tail())



# coding: utf-8

import tempfile
import six

if six.PY2:
    from urllib2 import urlopen, Request
else:
    from urllib.request import urlopen, Request


def demo_reporthook(downloaded, total_size):
    print("Downloaded {}, Total is {}".format(downloaded, total_size))


class BaseCralwer:

    def __init__(self, mode="http"):
        self.mode = "http"

    def fetch_and_parse(self, reporthook=None, path_to_download=None, proxies=None, chunksize=1024 * 50, *args,
                        **kwargs):
        """
        function to get data ,
        :param reporthook 使用urllib.request 的report_hook 来汇报下载进度 \
                    参考 https://docs.python.org/3/library/urllib.request.html#module-urllib.request
        :param path_to_download 数据文件下载的地址，如果没有提供，则下载到临时文件中，并在解析之后删除
        :param proxies urllib格式的代理服务器设置
        :return: 解析之后的数据结果
        """
        if (self.mode == "http"):
            download_file = self.fetch_via_http(reporthook=reporthook, path_to_download=path_to_download,
                                                proxies=proxies, chunksize=chunksize, *args, **kwargs)
        else:
            download_file = self.get_content(reporthook=reporthook, path_to_download=path_to_download,
                                             chunksize=chunksize, *args, **kwargs);

        result = self.parse(download_file, *args, **kwargs)
        try:
            download_file.close()
        except:
            pass
        return result

    def fetch_via_http(self, reporthook=None, path_to_download=None, proxies=None, chunksize=1024 * 50, *args,
                       **kwargs):
        if path_to_download is None:
            download_file = tempfile.NamedTemporaryFile(delete=True)
        else:
            download_file = open(path_to_download, 'wb')

        url = self.get_url(*args, **kwargs)

        request = Request(url)
        request.add_header('Referer', url)
        request.add_header('User-Agent',
                           r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.100 Safari/537.36")
        res = urlopen(request)

        if six.PY2:
            resinfo = res.info()
        else:
            resinfo = res

        if resinfo.getheader('Content-Length') is not None:
            total_size = int(resinfo.getheader('Content-Length').strip())
            downloaded = 0

            while True:
                chunk = res.read(chunksize)
                downloaded += len(chunk)
                if reporthook is not None:
                    reporthook(downloaded, total_size)
                if not chunk:
                    break
                download_file.write(chunk)
        else:
            content = res.read()
            download_file.write(content)

        download_file.seek(0)
        return download_file

    def get_url(self, *args, **kwargs):
        raise NotImplementedError("will impl in subclass")

    def get_content(self, reporthook=None, path_to_download=None, proxies=None, chunksize=1024 * 50, *args, **kwargs):
        raise NotImplementedError("will impl in subclass")

    def parse(self, download_file, *args, **kwargs):
        raise NotImplementedError("will impl in subclass")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jotdx',
 'jotdx.contrib',
 'jotdx.crawler',
 'jotdx.financial',
 'jotdx.parser',
 'jotdx.reader',
 'jotdx.tools',
 'jotdx.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'httpx>=0.20.0,<0.21.0',
 'loguru>=0.5.3,<0.6.0',
 'pandas>=1.3.4,<2.0.0',
 'prettytable>=2.2.1,<3.0.0',
 'py-mini-racer>=0.6.0,<0.7.0',
 'tenacity>=8.0.1,<9.0.0',
 'tqdm>=4.62.3,<5.0.0']

setup_kwargs = {
    'name': 'jotdx',
    'version': '0.1.4',
    'description': 'Get data',
    'long_description': None,
    'author': 'FangyangJz',
    'author_email': 'fangyang.jing@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
